/*  t h u r s d a y . c   justify text with really doovy spacing
 *  a  u  t  h  o  r  :   a   c   b
 *  c o m m e n c e d :   5-11-1994
 *
 *  This program is part of the Emperor Norton Utilities and is 
 *  freely distributable.
 */

#include <stdio.h>
#include <ctype.h>

#ifndef __GNUC__
#define inline /* mu */
#define volatile
#endif

/*
 *  u t i l i t y   f u n c t i o n s
 */

/* return number of whitespace characters in a string */

static inline int numwhite(char *frob) 
{
  int r=0;
  for(;(*frob);frob++) 
    r+=(isspace(*frob))?1:0;
  return r;
};

int max_width = 77;  /* maximum width to justify to */
int left_margin = 0; /* columns of whitespace to insert at left. */
int capitalise_runtogether = 1; /* BiCapitaliseTextIfItRunsTogether */
int random_pad = 1; /* randomly pad between words */
int break_width = -1; /* for user-specified line breaking width */

/* find the optimum intercharacter/interword spacing combination for a
 * string, returning it.
 */

int find_optimum(char *futplex, int *c, int *w, int *err)
{
  /* we evaluate all spacings from pure intercharacter to pure interword,
     and find the one which minimises the error value. */
  int error;
  int white = numwhite(futplex);
  int num_chars = strlen(futplex)-white;
  int pad_margin = max_width-num_chars; /* how much do we need to add? */
  int ci;

  error = pad_margin;

  /* iterate through all valid intercharacter spacings */
  for(ci=0; (ci*(num_chars-1))<=pad_margin; ci++) {
    register int wi;

    /* iterate through all additive interword spacings for this ic spacing */
    for(wi=0; (((wi*white)+(ci*(num_chars-1)))<=pad_margin); wi++) {
      int e=(pad_margin-((wi*white)+(ci*(num_chars-1))));
      /* the following test is biased towards less interword spacing and more
	 intercharacter spacing; to reverse, change the "<=" to a '<' */
      if((e<=error)&&(e>=0)) { /* we have a 'best' choice */
	/* the following if ensures that we have _some_ interword spacing
	   when there is no error incurred. */
	if((wi>0) || (e<error)) {
	  *c = ci;
	  *w = wi;
	  error = e;
	};
      };
      if(!white) break; /* break infinite loops */
    };
  };
  *err = error;
};

/* print a spaced string */

void print_spaced(char *foo, int chars, int words, int error)
{
  int last_space = 1;
  int wh = numwhite(foo);
  int i;
  int spacecount=wh;

  for(i=left_margin;i;i--) putchar(' ');
  for(;(*foo);foo++) {
    register int i;
    if(!isspace(*foo)) {
      putchar((last_space && capitalise_runtogether && (words<=chars) && wh)
	      ?toupper(*foo):*foo);
      last_space = 0;
    } else last_space = 1;
    if(*(foo+1)) {
      for(i=0; i<(isspace(*foo)?words:chars); i++) putchar(' ');
      /* if this is an inter-word gap and we have error to get rid of,
	 we do what we need, if we're Allowed to insert random padding. */
      if((isspace(*foo)) && random_pad)
	if((random()%(spacecount--))<error) {
	  putchar(' '); error--;
	};
    };
  };
/*  putchar('$');*/
  putchar('\n'); /* this will probably break MS-DOS compilers; hah suffer... */
};

volatile void usage(char *name)
{
  fprintf(stderr, "usage: %s [-w width] [-c] [-l left_margin]\n", name);
  exit(1);
};

main(int argc, char *argv[])
{
  char buffer[256];
  int i;

  /* read arguments */

  for(i=1;i<argc;i++) {
    if(*(argv[i])!='-') usage(argv[0]);
    switch(*(argv[i]+1)) {
    case 'w':
      max_width = atoi(argv[++i]);
      break;
    case 'c':
      capitalise_runtogether = 0;
      break;
    case 'l':
      left_margin = atoi(argv[++i]);
      break;
    case 'r':
      random_pad = 0;
      break;
    case 'b':
      break_width = atoi(argv[++i]);
      break;
    default:
      usage(argv[0]);
    };
  };

  /* set break width as need be */
  if ((break_width==-1)||(break_width>max_width)) break_width = max_width;

  /* t h e  m a i n  l o o p */

  while(!feof(stdin)) {
    int c, w, e;
    char buf2[256], *p;

    gets(buffer);
    if(!(*buffer)) continue;
    feof(stdin); /* kludge */
    if(feof(stdin)) break;

    /* first, check for the trivial case of one-character lines. */
    if(strlen(buffer)==1) {
      puts(buffer); continue;
    };

    /* handle buffer in fragments if length of input exceeds breaking width */
    p = buffer;

    while(strlen(p)>break_width) {
      char *q = p+break_width;
      char *r = NULL;
      register char *i;
      /* find last space before q */
      for(i=p; i<q; i++) if(isspace(*i)) r=i;
      if(r) {
	strncpy(buf2, p, r-p); buf2[r-p]='\0';
	find_optimum(buf2, &c, &w, &e); print_spaced(buf2, c, w, e); p=r+1;
      } else {
	strncpy(buf2, p, q-p-1); buf2[q-p-1]='\0';  strcat(buf2, "-");
	find_optimum(buf2, &c, &w, &e); print_spaced(buf2, c, w, e); p=q-1;
      };
    };

    find_optimum(p, &c, &w, &e);
    if(!feof(stdin))
      print_spaced(p, c, w, e);
  };
};
